-----------------------------------------------------------------------------------------------
-- Client Lua Script for PieMenu
-- @author daihenka
-----------------------------------------------------------------------------------------------
 
require "Apollo"
require "Window"

-----------------------------------------------------------------------------------------------
-- PieMenu Module Definitions
-----------------------------------------------------------------------------------------------
local ADDON_NAME     = "PieMenu"
local CONFIG_VERSION = 1

local PieMenu = Apollo.GetAddon(ADDON_NAME)
local L       = PieMenu.L
local DaiGUI  = Apollo.GetPackage("DaiGUI-1.0").tPackage
local DaiUtil = Apollo.GetPackage("DaiUtil-1.0").tPackage

local tostring, ipairs, pairs, Event_FireGenericEvent = tostring, ipairs, pairs, Event_FireGenericEvent

-----------------------------------------------------------------------------------------------
-- Constants and Local Variables
-----------------------------------------------------------------------------------------------
local tKeysDown = {}
local nPrevDown

-----------------------------------------------------------------------------------------------
-- UI Creation Functions
-----------------------------------------------------------------------------------------------
local function CreateTrackerForm(o)
  return DaiGUI:Create({
    AnchorFill         = true,
    Name               = "MouseTracker",
    Picture            = true,
    SwallowMouseClicks = false,
    Overlapped         = true,
    BGColor            = "88000000",
    Visible            = false,
    IgnoreMouse        = true,
    NoClip             = true,
    DoNotBlockTooltip  = true,
    Sprite             = "ClientSprites:WhiteFill",
    NewWindowDepth     = true,
    Events = {
      MouseMove = "OnMouseTrackerMouseMove"
    },
  }):GetInstance(o) -- , "TooltipStratum")
end
  

-----------------------------------------------------------------------------------------------
-- Core Addon Functions
-----------------------------------------------------------------------------------------------
function PieMenu:OnInitialize()
  -- slash commands
  Apollo.RegisterSlashCommand("piemenu", "OnSlashCmd", self)
  Apollo.RegisterEventHandler("SystemKeyDown",         "OnSystemKeyDown",         self)
  Apollo.RegisterTimerHandler("SystemKeyReleaseTimer", "OnSystemKeyReleaseTimer", self)
  Apollo.CreateTimer("SystemKeyReleaseTimer", 0.1, true)
    
  self.wndTracker = CreateTrackerForm(self)
  self.wndTracker:Show(false)
  
  for i = 1, #self.tMenus do
    if self.tMenus[i].wndTracker ~= self.wndTracker then
      self.tMenus[i].wndTracker = self.wndTracker
    end
  end
end

function PieMenu:OnEnable()
	self:RefreshUI()
end

function PieMenu:OnConfigure()
	self:ShowOptionsWindow()
end

-----------------------------------------------------------------------------------------------
-- PieMenu Settings
-----------------------------------------------------------------------------------------------
--- Addon save configuration event handler
-- @param eLevel addon save level
-- @return table with configuration data to save
function PieMenu:OnSaveSettings(eLevel)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Account then
    return
  end
	
	self.db.keybinds = {}
	for i, tMenu in ipairs(self.tMenus) do
		if tMenu.oKeybind then
			self.db.keybinds[tMenu.strId] = tostring(tMenu.oKeybind)
		else
			self.db.keybinds[tMenu.strId] = "--UNBOUND--"
		end
		
	end
  
  local tData = DaiUtil.TableCopy(self.db)
  tData.Version = CONFIG_VERSION
  
  DaiUtil.SerializeColors(tData)
  
  return tData
end



--- Addon restore configuration event handler
-- @param eLevel addon save level
-- @param tData saved addon configuration data
function PieMenu:OnRestoreSettings(eLevel, tData)
  if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Account then
    return
  end
  
  if (tData.Version or 0) < CONFIG_VERSION then
    -- discard old versions of the config as they are incompatible
    return
  end
  
  DaiUtil.DeserializeColors(tData)

  -- merge the settings with the defaults
  DaiUtil.TableMerge(self.db, tData)
	
	self:RefreshUI()
	
	if type(self.db.keybinds) == "table" then
		for i, tMenu in ipairs(self.tMenus) do
			local strKeybind = self.db.keybinds[tMenu.strId]
			if strKeybind == "--UNBOUND--" then
				tMenu.oKeybind = nil
			else
				tMenu.oKeybind = PieMenu.PieKey:new(strKeybind)
			end
		end
	end
end

function PieMenu:RefreshUI()
end


-----------------------------------------------------------------------------------------------
-- PieMenu Slash Commands
-----------------------------------------------------------------------------------------------
--- General slash command handler
-- @param cmd
-- @param args
function PieMenu:OnSlashCmd(cmd, args)
  self:ShowOptionsWindow()
end

local ktModifierKeys = { [16] = true, [17] = true, [18] = true }

function PieMenu:StopKeyBindMode()
	if self.bKeyBindMode then
		self.bKeyBindMode = false
		self.wndKeyBind:SetCheck(false)
		self.tKeyBindMenu = nil
		self.wndKeyBind = nil
		self.wndOptions:FindChild("KeybindBlocker"):Show(false)
		for idx, wndChild in ipairs(self.wndOptions:FindChild("KeybindContainer"):GetChildren()) do
			wndChild:FindChild("BindButton"):Enable(true)
		end
	end
end

function PieMenu:OnSystemKeyDown(nKey)
  -- not interested in modifier keys by themselves
  if ktModifierKeys[nKey] then return end
  
	
  local oKey = self.PieKey:new(nKey)
  local sKey = tostring(oKey)
	
	if self.bKeyBindMode and nKey ~= 27 then
		self.tKeyBindMenu.oKeybind = oKey
		self.wndKeyBind:SetText(tostring(oKey))
		self:StopKeyBindMode()
		return
	end
	
	
--  Print("nKey = " .. nKey .. " | " .. sKey)
  
  if tKeysDown[sKey] == nil then
		self:OnKeyDown(oKey)
    tKeysDown[sKey] = oKey
  end
  tKeysDown[sKey].nLastDown = Apollo.GetTickCount()
end


function PieMenu:OnSystemKeyReleaseTimer()
  local nCurrTime = Apollo.GetTickCount()
  local tExpiredKeys = {}
  
  for k, oKey in pairs(tKeysDown) do
     --[[
     After pressing down a key, the client always delays the second keydown event for that key
     by 300-500ms. This makes it impossible for a fine grain of control over the key state.
     ]]
    
    if (nCurrTime - oKey.nLastDown) >= 550 then -- we have released a key
      --Print(tostring(oKey) .. " has been released for " .. (nCurrTime - oKey.nLastDown) .. " ms")
			self:OnKeyUp(oKey)
      tKeysDown[k] = nil
    end
  end
end

function PieMenu:OnKeyDown(oKey)
  for idx, oMenu in ipairs(self.tMenus) do
    if oMenu ~= self.oCurrentMenu then
      if oMenu.oKeybind ~= nil and oMenu.oKeybind == oKey then
        self:CloseAllPieMenus()
        self.oCurrentMenu = oMenu
        self.oCurrentMenu:Show(true)
        self:UpdateMousePosition()
        return
      end
    end
  end
end

function PieMenu:OnKeyUp(oKey)
  if self.oCurrentMenu ~= nil and self.oCurrentMenu.oKeybind == oKey then
    self.oCurrentMenu:ActivateCurrentItem()
  end
end

function PieMenu:CloseAllPieMenus()
  for idx, oMenu in ipairs(self.tMenus) do
    oMenu:Show(false)
  end
  self.oCurrentMenu = nil
  self.wndTracker:DestroyChildren()
end

function PieMenu:UpdateMousePosition()
  if self.oCurrentMenu == nil or self.oCurrentMenu.wnd == nil then
    return
  end
  
  local tMouse = self.oCurrentMenu.wnd:FindChild("SliceContainer"):GetMouse()
  self.oCurrentMenu:DrawMouseLine(tMouse.x, tMouse.y)
  self.oCurrentMenu:SetCurrentItem(tMouse.x, tMouse.y)
end


function PieMenu:OnMouseTrackerMouseMove( wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY )
  self:UpdateMousePosition()
end