local P       = Apollo.GetAddon("PieMenu")
local DaiGUI  = Apollo.GetPackage("DaiGUI-1.0").tPackage
local DaiUtil = Apollo.GetPackage("DaiUtil-1.0").tPackage

P.PieMenu = {}
local PieMenu = P.PieMenu
local PieKey  = P.PieKey

local setmetatable, type, pairs, ipairs = setmetatable, type, pairs, ipairs

local knItemWidth       = 48
local knMinRadius       = 120
local knCenterOffset    = 15
local knOuterOffset     = 20
local kcrSegmentBorder  = "552ef6f4"
local kcrMouseLine      = "ff2ef6f4"
local HalfPI            = math.pi * 0.5
local PI2               = math.pi * 2

-----------------------------------------------------------------------------------------------
-- Utility/Helper functions
-----------------------------------------------------------------------------------------------

local function SetTargetMarker(markerIdx)
  local unitTarget = GameLib.GetTargetUnit()
  if unitTarget ~= nil and GroupLib.InGroup() then
    if unitTarget:GetTargetMarker() == markerIdx then
      unitTarget:ClearTargetMarker()
    else
      unitTarget:SetTargetMarker(markerIdx)
    end
  end
end

local function CalculateSegmentAngle(idx, nStep, nRotation, nBound)
  local nHalfStep = nStep / 2
  local nAngle = (idx-1) * nStep - nRotation
  
  if nBound == 1 then
    nAngle = nAngle - nHalfStep
  elseif nBound == 2 then
    nAngle = nAngle + nHalfStep
  end
  
  if nAngle > math.pi then
    nAngle = nAngle - PI2
  end

  return nAngle 
end

local function NormalizeAngle(nAngle)
  nAngle = (nAngle > 0) and (PI2 - nAngle) or (-nAngle)
	return nAngle
end

local function AngleBetween(nTarget, nAngle1, nAngle2)
  local nRAngle = ((nAngle2 - nAngle1) % PI2 + PI2) % PI2
  if nRAngle >= math.pi then
    nAngle2, nAngle1 = nAngle1, nAngle2
  end
  
  if nAngle1 <= nAngle2 then
    return nTarget >= nAngle1 and nTarget <= nAngle2
  else
    return nTarget >= nAngle1 or nTarget <= nAngle2
  end
end

local function GetWindowCenter(wnd)
  if wnd ~= nil then
    return wnd:GetWidth() / 2, wnd:GetHeight() / 2
  end
end

-----------------------------------------------------------------------------------------------
-- PieMenu
-----------------------------------------------------------------------------------------------
function PieMenu:new(o)
  o = o or {}
  setmetatable(o, self)
  self.__index = self
  
	if o.strId == nil then
		error("Menu requires an ID to be assigned at creation", 2)
	end
	
  o.strName = o.strName or "Unnamed Menu"
  o.nItemWidth = o.nItemWidth or knItemWidth
 
  -- convert string or table keybinds to a PieKey instance
  if not o.oKeybind and (o.strKeybind or o.tKeybind) then
    o.oKeybind = PieKey:new(o.strKeybind or o.tKeybind)
    o.strKeybind = nil
    o.tKeybind = nil
  end
  
  return o
end

-- Calculate the slice dimensions
function PieMenu:CalculateSlices()
  local nItemCount = #self.tItems
  local nStep = PI2 / nItemCount
  local nRotation = (90 / 180) * math.pi
  
  -- TODO: refactor the radius calculation to a function that will calculate a smooth radius 
  -- with adequite spacing between slices
  self.nRadius = self.nRadius or math.sqrt(self.nItemWidth) * math.sqrt(nItemCount) * 5.8
  if self.nRadius < knMinRadius then
    self.nRadius = knMinRadius
  end
  
  -- calculate item placements and angles
  for idx, tItem in ipairs(self.tItems) do
    tItem.leftAngle   = CalculateSegmentAngle(idx, nStep, nRotation, 1)
    tItem.rightAngle  = CalculateSegmentAngle(idx, nStep, nRotation, 2)
    tItem.centerAngle = CalculateSegmentAngle(idx, nStep, nRotation)
    
    tItem.x = math.cos(tItem.centerAngle) * self.nRadius - (self.nItemWidth / 2)
    tItem.y = math.sin(tItem.centerAngle) * self.nRadius - (self.nItemWidth / 2)
  end
  
  local nSide = (self.nRadius * 2) + self.nItemWidth
  self.wnd:SetAnchorOffsets(0, 0, nSide, nSide)
end


local function CreateMenuForm(o,wndParent)
  return DaiGUI:Create({
    Name                 = "Form",
    AnchorOffsets        = {10,10,517,517},
    SwallowMouseClicks   = true,
    Escapable            = true,
    Overlapped           = true,
    IgnoreMouse          = true,
    NoClip               = true,
    NewWindowDepth       = true,
    CloseOnExternalClick = true,
    Children = {
      { Name = "SliceContainer", AnchorFill = true },
      {
        Name          = "BGArt",
        AnchorPoints  = "FILL",
        AnchorOffsets = {-20,-20,20,20},
        Sprite        = "ClientSprites:WhiteCircle",
        BGColor       = "00000000",
        Picture       = true,
        IgnoreMouse   = true,
        NoClip        = true,
      },
      {
        Name            = "SliceText",
        AnchorCenter    = {160,80},
        Font            = "CRB_InterfaceLarge_BO",
        Sprite          = "CRB_Basekit:kitAccent_Glow_Blue",
        BGColor         = "ff888888",
        Picture         = true,
        DT_CENTER       = true,
        DT_VCENTER      = true,
        DT_WORDBREAK    = true,
        IgnoreMouse     = true,
        Visible         = true,
        NewControlDepth = 1500,
      },
      {
        Name            = "DebugText",
        AnchorPoints    = "CENTER",
        AnchorOffsets   = { -80, 42, 80, 122 },
        Font            = "CRB_InterfaceMedium_O",
        DT_CENTER       = true,
        DT_VCENTER      = true,
        DT_WORDBREAK    = true,
        Picture         = true,
        IgnoreMouse     = true,
        Visible         = true,
        NewControlDepth = 1499,
      },
      {
        Name            = "TrackerDebugText",
        AnchorPoints    = "CENTER",
        AnchorOffsets   = { -80, 102, 80, 132 },
        Font            = "CRB_InterfaceMedium_O",
        DT_CENTER       = true,
        DT_VCENTER      = true,
        DT_WORDBREAK    = true,
        Picture         = true,
        IgnoreMouse     = true,
        Visible         = true,
        NewControlDepth = 1499,
      },
    },
    Events = {
      MouseButtonDown = "OnWindowButtonDown",
      WindowClosed    = "OnWindowClose",
    },
  }):GetInstance(o, wndParent)
end



local function CreateItemButton(o, wndParent)
  return DaiGUI:Create({
    Name               = "ItemBtn",
    AnchorOffsets      = {0,0,48,48},
    BGColor            = "ChannelCommand",
    Font               = "Thick",
    SwallowMouseClicks = true,
    DT_VCENTER         = true,
    DT_CENTER          = true,
    Picture            = true,
    NoClip             = true,
    NewWindowDepth     = true,
    IgnoreMouse        = true,
    Children = {
      {
        Name              = "Highlight",
        AnchorPoints      = "FILL",
        AnchorOffsets     = {-10,-10,10,10},
        Picture           = true,
        IgnoreMouse       = true,
        NoClip            = true,
        NeverBringToFront = true,
        DoNotBlockTooltip = true,
      },
      {
        AnchorFill      = 2,
        Name            = "Icon",
        Picture         = true,
        IgnoreMouse     = true,
        NoClip          = true,
        NewControlDepth = 10,
      },
    },
  }):GetInstance(o, wndParent)
end



-- Show/Hide the menu
function PieMenu:Show(bShow)
  -- handle hiding the menu
  if bShow ~= nil and bShow == false then
    if self.wnd ~= nil and self.wnd:IsValid() then
      self.wnd:Close()
    end
    return
  end
  
  self.bDisabledMouseLock = GameLib.IsMouseLockOn()
  if self.bDisabledMouseLock then
    GameLib.SetMouseLock(false)
  end
    
  -- show the mouse tracker window
  self.wndTracker:Show(true)
  self.wndTracker:ToFront()
  
  -- handle showing the menu
  if self.wnd == nil or not self.wnd:IsValid() then
    self.wnd = CreateMenuForm(self, self.wndTracker)
    self.wnd:Show(false, true)
  end

  -- clear current item
  self.tCurrentItem = nil
  self:SetSliceText()
  
  
  -- populate the menu items if a function was supplied
  if self.fnPopulateItems ~= nil and type(self.fnPopulateItems) == "function" then
    self.tItems = self.fnPopulateItems()
  end
  
  -- calculate the slices
  self:CalculateSlices()
  
  
  local wndContainer = self.wnd:FindChild("SliceContainer")
  local tDisplaySize = Apollo.GetDisplaySize()
  local nCenterX, nCenterY = tDisplaySize.nWidth / 2, tDisplaySize.nHeight / 2
  local nMainX, nMainY = GetWindowCenter(self.wnd)
  
  self.wnd:SetAnchorOffsets(nCenterX - nMainX, nCenterY - nMainY, nCenterX + nMainX, nCenterY + nMainY)
  
  -- setup the mouse tracking line
  self.nMouseLinePixieId = self:DrawLine(wndContainer, {0,0,0,0}, kcrMouseLine, nil, 2)
  
  -- destroy children
  wndContainer:DestroyChildren()
  for idx, tItem in ipairs(self.tItems) do
    local wndCurr = CreateItemButton(self, wndContainer)
    if tItem.strIcon ~= nil then
      wndCurr:FindChild("Icon"):SetSprite(tItem.strIcon)
    else
      wndCurr:SetText(tItem.strName)
    end
    wndCurr:SetData(tItem)
    wndCurr:SetAnchorOffsets(nMainX + tItem.x, nMainY + tItem.y, 
                             nMainX + tItem.x + self.nItemWidth, nMainY + tItem.y + self.nItemWidth)
    wndCurr:Show(true)
    
    if #self.tItems > 1 then
      self:DrawSegmentBorder(tItem)
    end
  end
  
  self.wnd:Show(true, true)
  self.wnd:ToFront()
end

-- Call the function associated with the current menu item
function PieMenu:ActivateCurrentItem()
  self:ActivateItem(self.tCurrentItem)
end

-- Call the function associated with the item
-- @param tItem the menu item
function PieMenu:ActivateItem(tItem)
  if tItem ~= nil then
    if tItem.func ~= nil then
      tItem.func()
    elseif tItem.strEvent ~= nil then
      Event_FireGenericEvent(tItem.strEvent)
    elseif tItem.nMarkerId ~=  nil then
      SetTargetMarker(tItem.nMarkerId)
    elseif tItem.strInput ~= nil then
      Apollo.ParseInput(tItem.strInput)
    end
    
    
  end
  self:Show(false)
end

-- Highlight the current menu item
function PieMenu:HighlightCurrentItem()
  if self.tPrevItem == self.tCurrentItem then
    return
  end
  
  self:SetSliceText(self.tCurrentItem and self.tCurrentItem.strName or "")
  for idx, wndCurr in ipairs(self.wnd:FindChild("SliceContainer"):GetChildren()) do
    if wndCurr:GetData() == self.tCurrentItem then
      wndCurr:FindChild("Highlight"):SetSprite("PlayerPathContent_TEMP:spr_Crafting_TEMP_Stretch_QuestZonePulse")
    else
      wndCurr:FindChild("Highlight"):SetSprite()
    end
  end
  self.tPrevItem = self.tCurrentItem
end

-- Get the menu item at the X, Y coordinates
function PieMenu:GetMenuItemAt(x, y)
  -- if there is only 1 item, return it
  if #self.tItems == 1 then
    return self.tItems[1]
  else
    local wndControl = self.wnd:FindChild("SliceContainer")
    local nCX, nCY   = GetWindowCenter(wndControl)
    local nAngle     = NormalizeAngle(math.atan2(y - nCY, x - nCX))
    
    for idx, tItem in ipairs(self.tItems) do
      local nLeftBound  = NormalizeAngle(tItem.leftAngle)
      local nRightBound = NormalizeAngle(tItem.rightAngle)
      
      if AngleBetween(nAngle, nLeftBound, nRightBound) then
        -- found the segment
        return tItem
      end
    end
  end
end

-- Click on a menu item
function PieMenu:OnWindowButtonDown( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
  if wndHandler ~= wndControl then
    return
  end
  self:ActivateItem(self.tCurrentItem)
end

-- Set the current menu to the item located at x, y
function PieMenu:SetCurrentItem(x, y)
  self.tCurrentItem = self:GetMenuItemAt(x,y)
  self:HighlightCurrentItem()
end

-- Close the current menu window
function PieMenu:OnWindowClose()
  self.wndTracker:Show(false)
  self.tCurrentItem = nil
  if self.bDisabledMouseLock then
    GameLib.SetMouseLock(true)
  end
  
  self.bDisabledMouseLock = nil
  
  self.wnd:Destroy()
  if self.cbOnClose ~= nil then
    self.cbOnClose()
  end
end


function PieMenu:SetDebugText(strText)
  local wndControl = self.wnd:FindChild("DebugText")
  if wndControl ~= nil then
    wndControl:SetText(strText or "")
  end
end

function PieMenu:SetTrackerDebugText(strText)
  local wndControl = self.wnd:FindChild("TrackerDebugText")
  if wndControl ~= nil then
    wndControl:SetText(strText or "")
  end
end

function PieMenu:SetSliceText(strText)
  self.wnd:FindChild("SliceText"):SetText(strText)
end

-----------------------------------------------------------------------------------------------
-- Line Drawing Functions
-----------------------------------------------------------------------------------------------
function PieMenu:DrawMouseLine(nMouseX, nMouseY)
  local wnd = self.wnd:FindChild("SliceContainer")
  local nCX, nCY = GetWindowCenter(wnd)
  local nRadius = nCX - 50
  
  local nAngle = math.atan2(nMouseY - nCY, nMouseX - nCX)
  
  local nOffsets = { 
    nCX + knCenterOffset * math.cos(nAngle),
    nCY + knCenterOffset * math.sin(nAngle),
    nCX + nRadius * math.cos(nAngle),
    nCY + nRadius * math.sin(nAngle),
  }
  
  self:DrawLine(wnd, nOffsets, kcrMouseLine, self.nMouseLinePixieId)
end

function PieMenu:DrawSegmentBorder(tSegment)
  local wnd = self.wnd:FindChild("BGArt")
  local nRadius = self.wnd:FindChild("SliceContainer"):GetWidth() / 2
  local nCX, nCY = GetWindowCenter(wnd)
  self:DrawAngleLine(wnd, tSegment.leftAngle, nRadius)
  
--  local nCOffsets = { 
--    nCX + knCenterOffset * math.cos(tSegment.leftAngle),
--    nCY + knCenterOffset * math.sin(tSegment.leftAngle),
--    nCX + knCenterOffset * math.cos(tSegment.rightAngle),
--    nCY + knCenterOffset * math.sin(tSegment.rightAngle),
--  }
--  local nLOOffsets = { 
--    nCX + nRadius * math.cos(tSegment.leftAngle),
--    nCY + nRadius * math.sin(tSegment.leftAngle),
--    nCX + (nRadius + knOuterOffset) * math.cos(tSegment.centerAngle),
--    nCY + (nRadius + knOuterOffset) * math.sin(tSegment.centerAngle),
--  }
--  local nROOffsets = { 
--    nCX + (nRadius + knOuterOffset) * math.cos(tSegment.centerAngle),
--    nCY + (nRadius + knOuterOffset) * math.sin(tSegment.centerAngle),
--    nCX + nRadius * math.cos(tSegment.rightAngle),
--    nCY + nRadius * math.sin(tSegment.rightAngle),
--  }
 
--  self:DrawLine(wnd, nCOffsets, kcrSegmentBorder)
--  self:DrawLine(wnd, nLOOffsets, kcrSegmentBorder)
--  self:DrawLine(wnd, nROOffsets, kcrSegmentBorder)
end

function PieMenu:DrawAngleLine(wndControl, nAngle, nRadius)
  local nCX, nCY = GetWindowCenter(wndControl)
  nRadius = nRadius or nCX
  
  local nOffsets = { 
    nCX + knCenterOffset * math.cos(nAngle),
    nCY + knCenterOffset * math.sin(nAngle),
    nCX + nRadius * math.cos(nAngle),
    nCY + nRadius * math.sin(nAngle),
  }
  
  self:DrawLine(wndControl, nOffsets, kcrSegmentBorder)
end

function PieMenu:DrawLine(wndControl, nOffsets, oColor, nPixieId, fWidth)
  oColor = oColor or ApolloColor.new("9900ffff")
  fWidth = fWidth or 2
  
  local tPixieInfo = {
    bLine = true,
    fWidth = fWidth,
    cr = oColor,
    loc = {
      fPoints = {0,0,0,0},
      nOffsets = nOffsets,
    },
  }
  
  if nPixieId == nil or wndControl:GetPixieInfo(nPixieId) == nil then
    return wndControl:AddPixie(tPixieInfo)
  else
    wndControl:UpdatePixie(nPixieId, tPixieInfo)
    return nPixieId
  end
end

